from typing import Optional, List, Literal

import torch

from collie import trainer


class Callback:

    def on_train_start(
        self,
        trainer: "trainer.PytorchTrainer"
    ) -> None:
        ...

    def on_train_end(
        self,
        trainer: "trainer.PytorchTrainer"
    ) -> None:
        ...

    def on_epoch_start(
        self,
        epoch_step: int,
        trainer: "trainer.PytorchTrainer"
    ) -> None:
        ...

    def on_epoch_end(
        self, 
        epoch_step: int,
        epoch_train_loss: float,
        epoch_val_loss: Optional[float],
        trainer: "trainer.PytorchTrainer"
    ) -> None:
        ...
    
    def on_batch_start(
        self, 
        batch_step: int, 
        batch_data: torch.Tensor,
        trainer: "trainer.PytorchTrainer"
    ) -> None:
        ...

    def on_batch_end(
        self, 
        batch_step: int, 
        batch_data: torch.Tensor,
        batch_train_loss: float,
        trainer: "trainer.PytorchTrainer"
    ) -> None:
        ...


class _CallbackManager:
    def __init__(self, callbacks: List[Callback]):
        self.callbacks = callbacks

    def on_train_start(
        self,
        trainer: "trainer.PytorchTrainer"
    ) -> None:
        
        self._execute_callbacks(
            "on_train_start",
            trainer=trainer
        )
        
    def on_train_end(
        self,
        trainer: "trainer.PytorchTrainer"
    ) -> None:
        
        self._execute_callbacks(
            "on_train_end",
            trainer=trainer
        )

    def on_epoch_start(
        self, 
        trainer: "trainer.PytorchTrainer", 
        epoch_step: int
    ) -> None:

        self._execute_callbacks(
            "on_epoch_start",
            trainer=trainer,
            epoch_step=epoch_step
        )

    def on_epoch_end(
        self, 
        trainer: "trainer.PytorchTrainer", 
        epoch_step: int,
        epoch_train_loss: float,
        epoch_val_loss: Optional[float],
    ) -> None:
         
        self._execute_callbacks(
            "on_epoch_end",
            trainer=trainer,
            epoch_step=epoch_step,
            epoch_train_loss=epoch_train_loss,
            epoch_val_loss=epoch_val_loss
        )
         
    def on_batch_start(
        self,
        batch_step: int, 
        batch_data: torch.Tensor,
        trainer: "trainer.PytorchTrainer"
    ) -> None:
        
        self._execute_callbacks(
            "on_batch_start",
            trainer=trainer,
            batch_step=batch_step,
            batch_data= batch_data
        )

    def on_batch_end(
        self,
        batch_step: int, 
        batch_data: torch.Tensor,
        batch_train_loss: float,
        trainer: "trainer.PytorchTrainer"
    ) -> None:
        
        self._execute_callbacks(
            "on_batch_end",
            trainer=trainer,
            batch_step=batch_step,
            batch_data= batch_data,
            batch_train_loss=batch_train_loss
        )

    def _execute_callbacks(
        self, 
        method_name: Literal[
            "on_train_start",
            "on_train_end",
            "on_epoch_start", 
            "on_epoch_end", 
            "on_batch_start", 
            "on_batch_end"], 
        *args, **kwargs
    ) -> None:
       
        if self.callbacks is None:
            return
        
        for callback in self.callbacks:
            # Call the method specified by method_name on the callback
            getattr(callback, method_name)(*args, **kwargs)


from typing import (
    Optional,
    Dict, 
    Any
)
from abc import abstractmethod

from collie.contracts.event import Event, EventType
from collie.core.enums.components import CollieComponentType
from collie.contracts.mlflow import (
    MLFlowComponentABC,
    MLflowConfig
)
from collie._common.exceptions import (
    OrchestratorError,
    TrainerError,
    TunerError,
    EvaluatorError,
    PusherError,
    TransformerError,
)


class OrchestratorBase(MLFlowComponentABC):

    def __init__(
        self,
        components: CollieComponentType,
        tracking_uri: Optional[str] = None,
        mlflow_tags: Optional[Dict[str, str]] = None,
        experiment_name: Optional[str] = None,
        description: Optional[str] = None
    ) -> None:

        super().__init__()
        self.components = components
        self.mlflow_tags = mlflow_tags
        self.tracking_uri = tracking_uri
        self.description = description
        self.experiment_name = experiment_name

    @abstractmethod
    def orchestrate_pipeline(self) -> Any:
        raise NotImplementedError

    def run(self) -> Any:
        
        self.mlflow_config = MLflowConfig(
            tracking_uri=self.tracking_uri,
            experiment_name=self.experiment_name,
        )
        try:
            with self.start_run(
                tags=self.mlflow_tags, 
                run_name="Orchestrator", 
                description=self.description, 
            ):
                return self.orchestrate_pipeline()
        except (
            TrainerError,
            TunerError,
            EvaluatorError,
            PusherError,
            TransformerError,
        ) as e:
            raise OrchestratorError(
                f"Component error in orchestration: {str(e)}"
            ) from e
        except Exception as e:
            raise OrchestratorError(
                f"Unexpected orchestration error: {str(e)}"
            ) from e

    def initialize_event(self) -> Event:
        """Initialize pipeline with an event."""
        return Event(
            type=EventType.INITIALIZE,
            payload=None
        )


from typing import Union
from enum import Enum

from collie.core.transform.transform import Transformer
from collie.core.tuner.tuner import Tuner
from collie.core.trainer.trainer import Trainer
from collie.core.evaluator.evaluator import Evaluator
from collie.core.pusher.pusher import Pusher


class CollieComponents(Enum):

    TRAINER = Trainer
    TRANSFORMER = Transformer
    TUNER = Tuner 
    EVALUATOR = Evaluator
    PUSHER = Pusher


CollieComponentType = Union[
    Trainer, 
    Transformer, 
    Tuner,
    Evaluator, 
    Pusher
]


from typing import Any, Optional

import mlflow
from mlflow.tracking import MlflowClient
from collie._common.mlflow_model_io.flavor_registry import FlavorRegistry


class MLflowModelIO:
    def __init__(
        self, 
        mlflow_client: MlflowClient
    ) -> None:
        """
        Initializes an MLflowModelIO instance.

        Args:
            mlflow_client (MlflowClient): The MLflowClient instance to use for logging models.

        """
        self.registry = FlavorRegistry()
        self.client = mlflow_client

    def log_model(
        self,
        model: Any,
        name: str,
        registered_model_name: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        """
        Logs a model with MLflow.

        Args:
            model (Any): The model to log.
            name (str): The name to give the logged model.
            registered_model_name (Optional[str], optional): The name to give the registered model. Defaults to None.
            **kwargs (Any): Additional keyword arguments to pass to the flavor handler's log_model method.

        Raises:
            ValueError: If the model type is not supported by any flavor handler.

        """
        handler = self.registry.find_handler_by_model(model)
        if handler is None:
            raise ValueError(f"Unsupported model type: {type(model)}")

        handler.log_model(
            model=model, 
            name=name, 
            registered_model_name=registered_model_name, 
            **kwargs
        )
        mlflow.log_param("model_flavor", handler.flavor())

    def load_model(
        self, 
        flavor: str,
        model_uri: str,
    ) -> Any:

        handler = self.registry.find_handler_by_flavor(flavor)
        if handler is None:
            raise ValueError(f"Unsupported model flavor: {flavor}")

        return handler.load_model(model_uri)


from typing import Optional
from collie.contracts.event import (
    Event, 
    EventHandler, 
    EventType
)
from collie.contracts.mlflow import MLFlowComponentABC
from collie.core.models import PusherPayload
from collie._common.decorator import type_checker
from collie._common.exceptions import PusherError
from collie.core.enums.ml_models import MLflowModelStage


class Pusher(EventHandler, MLFlowComponentABC):
    def __init__(
        self,
        target_stage: MLflowModelStage = MLflowModelStage.PRODUCTION,
        archive_existing_versions: bool = True,
        description: Optional[str] = None,
        tags: Optional[dict] = None
    ) -> None:
        """
        Initializes the Pusher.

        Args:
            target_stage (Optional[MLflowModelStage], optional): The stage to transition the model to after evaluation. Defaults to None.
            archive_existing_versions (bool, optional): Whether to archive existing versions at the target stage. Defaults to True.
            description (Optional[str], optional): Description for the MLflow run. Defaults to None.
            tags (Optional[dict], optional): Tags to associate with the MLflow run. Defaults to None.

        """
        super().__init__()
        self._registered_model_name = None
        self.target_stage = target_stage
        self.archive_existing_versions = archive_existing_versions
        self.description = description
        self.tags = tags or {"component": "Pusher"}
    
    @property
    def registered_model_name(self) -> str:
        if not self._registered_model_name:
            raise PusherError("Registered model name is not set.")
        return self._registered_model_name
    
    @registered_model_name.setter
    def registered_model_name(self, name: str) -> None:
        self._registered_model_name = name

    def run(self, event: Event) -> Event:
        with self.start_run(
            tags=self.tags,
            run_name="Pusher",
            log_system_metrics=False,
            nested=True,
            description=self.description
        ):
            try:
                pusher_event = self._handle(event)
                payload = self._get_pusher_payload(pusher_event)
                pass_evaluation = event.context.get("pass_evaluation")
                
                if pass_evaluation:
                    model_uri = event.context.get('model_uri')
                    if not model_uri:
                        raise PusherError("model_uri not found in event context")
                    
                    version = self.register_model(
                        model_name=self.registered_model_name,
                        model_uri=model_uri
                    )
                    self.mlflow.log_param("model_registered", "true")
                    event.context.set("registered_version", str(version))
                    self.mlflow.log_param("model_version", version)
                    
                    self.mlflow.log_param("target_stage", self.target_stage.value)
                    self.transition_model_version(
                        registered_model_name=self.registered_model_name,
                        version=str(version),
                        desired_stage=self.target_stage,
                        archive_existing_versions_at_stage=self.archive_existing_versions,
                    )

                else:
                    self.mlflow.log_param("skip_reason", "evaluation_failed")
                    self.mlflow.log_param("model_registered", "false")

                return Event(
                    type=EventType.PUSHER_DONE,
                    payload=payload,
                    context=event.context,
                )

            except Exception as e:
                raise PusherError(f"Pusher failed with error: {e}")

    @type_checker((PusherPayload,), "PusherPayload must be of type PusherPayload.")
    def _get_pusher_payload(self, event: Event) -> PusherPayload:
        return event.payload


from typing import (
    Dict, 
    Any,
    Optional,
    List,
)
from pydantic import BaseModel, ConfigDict, Field

import pandas as pd


class BasePayload(BaseModel):
    """Base class for all payload types with optional extra_data functionality."""
    
    extra_data: Optional[Dict[str, Any]] = None
    
    model_config = ConfigDict(arbitrary_types_allowed=True)
    
    def set_extra(self, key: str, value: Any):
        """
        Set a value in extra_data and return self for chaining.
        
        Args:
            key: The key to set
            value: The value to store
            
        Returns:
            Self for method chaining
            
        Example:
            >>> payload.set_extra("feature_names", ["age", "income"])
            >>> payload.set_extra("n_classes", 3).set_extra("version", "1.0")
        """
        if self.extra_data is None:
            self.extra_data = {}
        self.extra_data[key] = value
        return self
    
    def get_extra(self, key: str, default: Any = None) -> Any:
        """
        Get a value from extra_data with optional default.
        
        Args:
            key: The key to retrieve
            default: Default value if key doesn't exist
            
        Returns:
            The value or default
            
        Example:
            >>> features = payload.get_extra("feature_names", [])
        """
        if self.extra_data is None:
            return default
        return self.extra_data.get(key, default)
    
    def has_extra(self, key: str) -> bool:
        """
        Check if a key exists in extra_data.
        
        Args:
            key: The key to check
            
        Returns:
            True if key exists, False otherwise
            
        Example:
            >>> if payload.has_extra("feature_names"):
            ...     features = payload.get_extra("feature_names")
        """
        if self.extra_data is None:
            return False
        return key in self.extra_data


class TransformerArtifact(BaseModel):
    train_data: str = "train.csv"
    validation_data: str = "validation.csv"
    test_data: str = "test.csv"


class TunerArtifact(BaseModel):
    hyperparameters: str = "hyperparameters.json"


class TrainerArtifact(BaseModel):
    model: str = "model"


class EvaluatorArtifact(BaseModel):
    report: str = "report.json"


class TransformerPayload(BasePayload):
    train_data: Optional[pd.DataFrame] = None
    validation_data: Optional[pd.DataFrame]  = None
    test_data: Optional[pd.DataFrame]  = None

    model_config = ConfigDict(arbitrary_types_allowed=True)


class TrainerPayload(BasePayload):
    model: Any = None

    model_config = ConfigDict(arbitrary_types_allowed=True)


class TunerPayload(BasePayload):
    hyperparameters: Dict[str, Any]
    train_data: Optional[pd.DataFrame] = None
    validation_data: Optional[pd.DataFrame]  = None
    test_data: Optional[pd.DataFrame]  = None

    model_config = ConfigDict(arbitrary_types_allowed=True)


class EvaluatorPayload(BasePayload):
    metrics: List[Dict[str, Any]]
    is_better_than_production: bool


class PusherPayload(BasePayload):
    model_uri: str
    status: Optional[str] = None
    model_version: Optional[str] = None


from typing import Optional
from collie.contracts.event import (
    Event, 
    EventHandler, 
    EventType
)
from collie.contracts.mlflow import MLFlowComponentABC
from collie._common.decorator import type_checker
from collie.core.models import (
    TunerArtifact,
    TunerPayload
)
from collie._common.exceptions import TunerError


class Tuner(EventHandler, MLFlowComponentABC):

    def __init__(
        self,
        description: Optional[str] = None,
        tags: Optional[dict] = None
    ) -> None:
        """
        Initializes the Tuner component.

        Args:
            description (Optional[str], optional): Description for the MLflow run. Defaults to None.
            tags (Optional[dict], optional): Tags to associate with the MLflow run. Defaults to None.
        """
        super().__init__()
        self.description = description
        self.tags = tags or {"component": "Tuner"}
    
    def run(self, event: Event) -> None:
        """
        Run the hyperparameter tuner component.

        This method starts a new MLflow run, tunes the hyperparameters,
        logs metrics, and sets the outputs.
        """

        with self.start_run(
            tags=self.tags,
            run_name="Tuner",
            log_system_metrics=True,
            nested=True,
            description=self.description
        ) as run:
            try:
                tuner_event = self._handle(event)

                tuner_payload = self._tuner_payload(tuner_event)
                hyperparameters = tuner_payload.model_dump()

                self.mlflow.log_dict(
                    dictionary=hyperparameters, 
                    artifact_file=TunerArtifact().hyperparameters
                )
                event.context.set(
                    "hyperparameters_uri",
                    self.artifact_path(run)
                )

                event_type = EventType.TUNING_DONE

                return Event(
                    type=event_type,
                    payload=tuner_payload,
                    context=event.context
                )
            except Exception as e:
                raise TunerError(f"Tuner failed with error: {e}") from e

    @type_checker((TunerPayload,) , 
        "TunerPayload must be of type TunerPayload."
    )    
    def _tuner_payload(self, event: Event) -> TunerPayload: 

        tuner_payload: TunerPayload = event.payload

        return tuner_payload
    
    def artifact_path(self, run) -> str:
        return f"{run.info.artifact_uri}/{TunerArtifact().hyperparameters}"


from typing import (
    Optional, 
    Tuple, 
    Dict,
    Any,
    List,
)
from abc import abstractmethod


from tqdm import tqdm
from torch.utils.data import DataLoader
from torch import nn
from torch.amp import GradScaler, autocast
import torch

from collie.trainer.trainer import Trainer
from collie._common.decorator import type_checker
from collie._common.utils import get_logger
from collie.helper.pytorch.callback.callback import Callback, _CallbackManager
from collie.helper.pytorch.callback.model_checkpoint import ModelCheckpoint
from collie.helper.pytorch.callback.earlystop import EarlyStopping
from collie.contracts.event import Event
from collie.core.types import TrainerPayload, TransformerPayload


#TODO: Features to develop:
# 1. Train with multiple GPUs 
logger = get_logger()


class _AbstractPytorchTrainer:
    
    @abstractmethod
    def configure_optimizers(
        self
    ) -> Tuple[
        torch.optim.Optimizer, 
        Optional[torch.optim.lr_scheduler._LRScheduler]
        ]:
        """
        Configure the optimizers and learning rate schedulers.

        Returns:
            A tuple of two objects. The first object is the optimizers and the second is the learning rate schedulers if necessary.

        Example:
            code block :: python

            def configure_optimizers(self) -> Tuple[List[torch.optim.Optimizer], List[torch.optim.lr_scheduler._LRScheduler]]:
                optimizer = torch.optim.AdamW(self.model.parameters(), lr=self.initial_lr)
                scheduler = get_cosine_schedule_with_warmup(
                    optimizer=optimizer,
                    num_warmup_steps=self.num_warmup_steps,
                    num_training_steps=self.num_training_steps
                )

                return optimizer, scheduler
        """
        raise NotImplementedError("Please implement the *configure_optimizers* method.")
    
    @type_checker(
        (DataLoader,),
        "The train_data in TransformerPayload should be type of DataLoader in PytorchTrainer."
    )
    def get_train_dataloader(self, event: Event[TransformerPayload]) -> DataLoader:
        transformer_payload = event.payload
        return transformer_payload.train_data

    @type_checker(
        (DataLoader, None),
        "The validation_data in TransformerPayload should be type of DataLoader in PytorchTrainer."
    )
    def get_val_dataloader(self, event: Event[TransformerPayload]) -> DataLoader:
        transformer_payload = event.payload
        return transformer_payload.validation_data


class PytorchTrainer(_AbstractPytorchTrainer):
    def __init__(
        self, 
        model: nn.Module,
        epochs: int,
        device: Optional[str],
        use_amp: Optional[bool],
        topk_checkpoints: Optional[int],
        earlystop_patience_on_epoch: Optional[int] = None,
        accumulate_grad_batches: int = 1,
        callbacks: Optional[List[Callback]] = None,
    ) -> None:

        super().__init__()

        self.model = model

        self.epochs = epochs
        self.device = device or ("cuda" if torch.cuda.is_available() else "cpu")
        self.use_amp = use_amp
        self.topk_checkpoints = 3 if topk_checkpoints is None else topk_checkpoints
        self.earlystop_patience_on_epoch = earlystop_patience_on_epoch

        DEFAULT_CALLBACK = [ModelCheckpoint(topk_checkpoints=self.topk_checkpoints)]
        if self.earlystop_patience_on_epoch:
            DEFAULT_CALLBACK.append(EarlyStopping(self.earlystop_patience_on_epoch, delta=0.0))
        
        self.callbacks = [] if callbacks is None else callbacks
        self.callbacks = self.callbacks + DEFAULT_CALLBACK
        self.accumulate_grad_batches = accumulate_grad_batches

        self.train_data_loader = None
        self.val_data_loader = None
        self.optimizer = None
        self.lr_scheduler = None
        self.grad_scaler = None
        self.should_stop = False
        self.cb_manager = _CallbackManager(callbacks=self.callbacks)

        self.model = self.model.to(self.device)

    def run(
        self, 
        event: Event[TransformerPayload]
    ) -> Event[TrainerPayload]:

        self.train_data_loader = self.get_train_dataloader(event=event)
        self.val_data_loader = self.get_val_dataloader(event=event)

        self.optimizer, self.lr_scheduler = self.configure_optimizers()
        
        trainstep_per_epoch = len(self.train_data_loader)
        total_train_step = trainstep_per_epoch * self.epochs

        self.log_param("epoch", self.epochs)
        self.log_param("batch size", self.train_data_loader.batch_size)
        self.log_param("total training step", total_train_step)

        self.cb_manager.on_train_start(self)
        self.optimizer.zero_grad()
        # TODO: the train loop is written by the user in handle method.
        for epoch_idx in tqdm(range(1, self.epochs + 1)):
            # Modified by Earlystop callback
            if self.should_stop:
                break

            self.cb_manager.on_epoch_start(self, epoch_idx)
            running_loss = 0
            self.model.train()
            for batch_idx, batch_data in enumerate(self.train_data_loader, start=1):
                self.cb_manager.on_batch_start(
                    batch_data=batch_data,
                    batch_step=batch_idx,
                    trainer=self
                )
                
                loss = self._get_train_step_result(epoch_idx, batch_data) / self.accumulate_grad_batches
                running_loss += loss.item()  * self.accumulate_grad_batches

                self._backward(loss=loss, batch_idx=batch_idx)
                # The learning rate scheduler is batch level.
                if self.lr_scheduler and batch_idx % self.accumulate_grad_batches == 0:
                    self.lr_scheduler.step()

                self.cb_manager.on_batch_end(
                    batch_step=batch_idx,
                    batch_data=batch_data,
                    batch_train_loss=running_loss,
                    trainer=self
                )

            epoch_loss = running_loss / trainstep_per_epoch
            
            self.log_metric("train loss", epoch_loss, step=epoch_idx)
            self._log_lr(epoch_step=epoch_idx)

            epoch_val_loss = self.validation_loop(epoch_step=epoch_idx)
            
            self.cb_manager.on_epoch_end(
                trainer=self,
                epoch_step=epoch_idx,
                epoch_train_loss=epoch_loss,
                epoch_val_loss=epoch_val_loss
            )
        
        # TODO: Fix the issue of loading the model from the following result model.
        self.log_model(self.model, model_type="pt")

        self.cb_manager.on_train_end(self)

        # return {"model":self.model, "loss": epoch_loss}
        trainer_payload = TrainerPayload(
            model=self.model,
            train_loss=epoch_loss,
            val_loss=epoch_val_loss
        )
        return Event(
            payload=trainer_payload
        )

    def validation_loop(self, epoch_step: int) -> Optional[float]:
        
        if self.val_data_loader is None:
            return None
        
        valstep_per_epoch = len(self.val_data_loader)

        self.model.eval()
        val_running_loss = 0
        with torch.no_grad():
            
            for val_batch_data in self.val_data_loader:
                
                val_loss = self.validation_step(epoch_step, val_batch_data)
                val_running_loss += val_loss.item()

            epoch_val_loss = val_running_loss / valstep_per_epoch
        
        self.log_metric("val_loss", epoch_val_loss, step=epoch_step)
        return epoch_val_loss

    def _backward(
        self, 
        loss: torch.Tensor, 
        batch_idx: int
    ) -> None:

        def should_step_optimizer(batch_idx: int) -> bool:
            return batch_idx % self.accumulate_grad_batches == 0 or batch_idx == len(self.train_data_loader)

        if self.use_amp:
            if self.grad_scaler is None:
                self.grad_scaler = GradScaler(device=self.device)

            self.grad_scaler.scale(loss).backward()
            if should_step_optimizer(batch_idx):
                
                if self._has_invalid_gradients():
                    # gradient clipping
                    self.grad_scaler.unscale_(self.optimizer)
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_norm=1.0)
                
                self.grad_scaler.step(self.optimizer)  
                self.grad_scaler.update() 
                self.optimizer.zero_grad()
        else:
            loss.backward()
            if should_step_optimizer():
                if self._has_invalid_gradients(batch_idx):
                    # gradient clipping
                    torch.nn.utils.clip_grad_norm_(self.model.parameters(), max_norm=1.0)
                self.optimizer.step()
                self.optimizer.zero_grad()
    
    def _has_invalid_gradients(self) -> bool:
   
        for param in self.model.parameters():
            if param.grad is not None:
                if torch.isnan(param.grad).any() or torch.isinf(param.grad).any():
                    logger.warning(f"Detected NaN or Inf gradients for parameter {param}. Zeroing out the gradients.")
                    # side effect
                    param.grad.zero_()
                    return True
        return False
            
    def _log_lr(self, epoch_step: int) -> None:
        
        lr = self.optimizer.param_groups[0]['lr'] if not self.lr_scheduler else self.lr_scheduler.get_last_lr()[0]
        self.log_metric("learning rate", lr, step=epoch_step)
 
    @type_checker((torch.Tensor,), 
                  "The return type of *train_step* method must be 'Tensor'."
    )
    def _get_train_step_result(
        self, 
        epoch_step: int, 
        batch_data: torch.Tensor
    ):

        if self.use_amp:
            with autocast(device_type=self.device):
                loss = self.handle(epoch_step, batch_data)
        else:
            loss = self.handle(epoch_step, batch_data)

        return loss
    
    def _get_optimizers(
        self
    ) -> Tuple[torch.optim.Optimizer, Optional[torch.optim.lr_scheduler.LRScheduler]]:
        """
        Get the optimizer and learning rate scheduler from the 
        *configure_optimizers* method.
        """
        result = self.configure_optimizers()
        if isinstance(result, tuple):
            optimizers, lr_scheduler = result
            if not isinstance(lr_scheduler, torch.optim.lr_scheduler._LRScheduler):
                raise TypeError(
                    f"learning rate scheduler must be type of \
                    torch.optim.lr_scheduler._LRScheduler "
                )
        else:
            optimizers, lr_scheduler = result, None
        return optimizers, lr_scheduler
        


import logging


def get_logger() -> logging.Logger:
    """
    Return a logger that logs messages with severity level info or higher.
    
    Returns:
        A logger that logs messages with severity level info or higher.
    """
    logging.basicConfig(level=logging.INFO)  
    logger = logging.getLogger(__name__)
    return logger



"""
Collie - A Lightweight MLOps Framework for Machine Learning Workflows

Collie provides a modular, event-driven architecture for building ML pipelines
with deep MLflow integration.

Quick Start:
    >>> from collie import Transformer, Trainer, Orchestrator
    >>> # Define your components
    >>> orchestrator = Orchestrator(
    ...     components=[MyTransformer(), MyTrainer()],
    ...     tracking_uri="http://localhost:5000",
    ...     registered_model_name="my_model"
    ... )
    >>> orchestrator.run()

For more examples, see: https://github.com/ChingHuanChiu/collie
"""

__author__ = "ChingHuanChiu"
__email__ = "stevenchiou8@gmail.com"
__version__ = "0.1.0b0"

# Import all main components for easy access
from .contracts.event import Event, EventType, PipelineContext
from .core.transform.transform import Transformer
from .core.trainer.trainer import Trainer
from .core.tuner.tuner import Tuner
from .core.evaluator.evaluator import Evaluator
from .core.pusher.pusher import Pusher
from .core.orchestrator.orchestrator import Orchestrator

# Import data models
from .core.models import (
    TransformerPayload,
    TrainerPayload,
    TunerPayload,
    EvaluatorPayload,
    PusherPayload,
)

# Import enums for configuration
from .core.enums.ml_models import ModelFlavor, MLflowModelStage

__all__ = [
    # Core components - the main classes users interact with
    "Transformer",
    "Trainer",
    "Tuner",
    "Evaluator",
    "Pusher",
    "Orchestrator",
    
    # Event system - for building custom workflows
    "Event",
    "EventType",
    "PipelineContext",
    
    # Payload models - for type-safe data passing
    "TransformerPayload",
    "TrainerPayload",
    "TunerPayload",
    "EvaluatorPayload",
    "PusherPayload",
    
    # Configuration enums
    "ModelFlavor",
    "MLflowModelStage",
]


from typing import Tuple, List
from functools import wraps


def type_checker(
    typing: Tuple[type], 
    error_msg: str
):
    """
    A decorator that checks the type of the output of a function.

    Args:
        typing (Tuple[type]): A tuple of types to check against.
        error_msg (str): The error message to be raised if the type does not match.

    Raises:
        TypeError: If the type of the output of the function does not match with given types.
    """
    
    def closure(func):
        @wraps(func)
        def wrapper(*arg, **kwarg):
            result = func(*arg, **kwarg)
            if not isinstance(result, typing):
                raise TypeError(error_msg)
            return result
        return wrapper
    return closure


def dict_key_checker(keys: List[str]):
    """
    A decorator that checks the keys of the output of a function.

    Args:
        keys (List[str]): A list of keys to check against.

    Raises:
        TypeError: If the output of the function is not a dictionary.
        KeyError: If the output of the function does not contain all the keys in the list.
    """
    def closure(func):
        @wraps(func)
        def wrapper(*arg, **kwarg):
            result = func(*arg, **kwarg)
            if not isinstance(result, dict):
                raise TypeError("The output must be a dictionary.")
            all_keys_exist = all(key in result for key in keys)
            if not all_keys_exist:
                raise KeyError(f"The following keys must all exist in the output: {keys}. Output: {result}")
            return result
        return wrapper
    return closure


from typing import Optional
from collie.contracts.event import (
    Event, 
    EventHandler, 
    EventType
)
from collie.contracts.mlflow import MLFlowComponentABC
from collie.core.models import (
    TrainerPayload,
    TrainerArtifact
)
from collie._common.decorator import type_checker
from collie._common.exceptions import TrainerError


class Trainer(EventHandler, MLFlowComponentABC):

    def __init__(
        self, 
        description: Optional[str] = None,
        tags: Optional[dict] = None
    ) -> None:
        """
        Initializes the Trainer component.

        Args:
            description (Optional[str], optional): Description for the MLflow run. Defaults to None.
            tags (Optional[dict], optional): Tags to associate with the MLflow run. Defaults to None.
        """
        super().__init__()
        self.description = description
        self.tags = tags or {"component": "Trainer"}
    
    def run(self, event: Event) -> Event:
        """
        Run the trainer component.

        This method starts a new MLflow run, trains the model,
        logs metrics, and sets the outputs.
        """
        
        with self.start_run(
            run_name="Trainer", 
            tags=self.tags,
            log_system_metrics=True, 
            nested=True,
            description=self.description
        ) as run:
            try:
                trainer_event = self._handle(event)

                trainer_payload = self._trainer_payload(trainer_event)
                event_type = EventType.TRAINING_DONE
                
                model = trainer_payload.model
                self.log_model(
                    model=model, 
                    name=TrainerArtifact().model
                )
                event.context.set("model_uri", self.artifact_uri(run))

                return Event(
                    type=event_type,
                    payload=trainer_payload,
                    context=event.context
                )
            except Exception as e:
                raise TrainerError(f"Trainer failed with error: {e}") from e

    @type_checker((TrainerPayload,) , 
        "TrainerPayload must be of type TrainerPayload."
    )    
    def _trainer_payload(self, event: Event) -> TrainerPayload: 

        return event.payload
    
    def artifact_uri(self, run) -> str:
        return f"runs:/{run.info.run_id}/{TrainerArtifact().model}"


from typing import List, Optional

from collie._common.mlflow_model_io.base_flavor_handler import FlavorHandler
from collie._common.mlflow_model_io.model_flavors import (
    SklearnFlavorHandler,
    XGBoostFlavorHandler, 
    PyTorchFlavorHandler,
    LightGBMFlavorHandler,
    TransformersFlavorHandler,
    SKLEARN_AVAILABLE,
    XGBOOST_AVAILABLE,
    PYTORCH_AVAILABLE,
    LIGHTGBM_AVAILABLE,
    TRANSFORMERS_AVAILABLE
)
from collie._common.exceptions import ModelFlavorError


class FlavorRegistry:
    """Registry for model flavor handlers with conditional loading."""
    
    def __init__(self):
        self._handlers: List[FlavorHandler] = []
        
        # Only register handlers for available frameworks
        if SKLEARN_AVAILABLE:
            self._handlers.append(SklearnFlavorHandler())
        if XGBOOST_AVAILABLE:
            self._handlers.append(XGBoostFlavorHandler())
        if PYTORCH_AVAILABLE:
            self._handlers.append(PyTorchFlavorHandler())
        if LIGHTGBM_AVAILABLE:
            self._handlers.append(LightGBMFlavorHandler())
        if TRANSFORMERS_AVAILABLE:
            self._handlers.append(TransformersFlavorHandler())
            
        if not self._handlers:
            raise ModelFlavorError(
                "No model flavor handlers available. Please install at least one supported ML framework."
            )

    def find_handler_by_model(self, model) -> Optional[FlavorHandler]:
        """Find a handler that can handle the given model."""
        for handler in self._handlers:
            if handler.can_handle(model):
                return handler
        return None

    def find_handler_by_flavor(self, flavor: str) -> Optional[FlavorHandler]:
        """Find a handler by flavor name."""
        for handler in self._handlers:
            if handler.flavor() == flavor:
                return handler
        return None
    
    def get_available_flavors(self) -> List[str]:
        """Get list of available model flavors."""
        return [handler.flavor().value for handler in self._handlers]
    
    def get_handler_info(self) -> dict:
        """Get information about registered handlers."""
        return {
            "total_handlers": len(self._handlers),
            "available_flavors": self.get_available_flavors(),
            "framework_status": {
                "sklearn": SKLEARN_AVAILABLE,
                "xgboost": XGBOOST_AVAILABLE,
                "pytorch": PYTORCH_AVAILABLE,
                "lightgbm": LIGHTGBM_AVAILABLE,
                "transformers": TRANSFORMERS_AVAILABLE
            }
        }


from .transform.transform import Transformer
from .trainer.trainer import Trainer
from .tuner.tuner import Tuner
from .evaluator.evaluator import Evaluator
from .pusher.pusher import Pusher
from .orchestrator.orchestrator import Orchestrator
from .models import (
    TransformerPayload,
    TrainerPayload,
    TunerPayload,
    EvaluatorPayload,
    PusherPayload,
    TrainerArtifact,
    TransformerArtifact,
    TunerArtifact,
    EvaluatorArtifact
)
from .enums.components import CollieComponentType, CollieComponents
from .enums.ml_models import ModelFlavor


__all__ = [
    "Transformer",
    "Trainer",
    "Tuner",
    "Evaluator",
    "Pusher",
    "Orchestrator",
    "TransformerPayload",
    "TrainerPayload",
    "TunerPayload",
    "EvaluatorPayload",
    "PusherPayload",
    "CollieComponentType",
    "CollieComponents",
    "TrainerArtifact",
    "TransformerArtifact",
    "TunerArtifact",
    "EvaluatorArtifact",
    "ModelFlavor"
]


from typing import Optional
from collie.contracts.event import (
    Event, 
    EventHandler, 
    EventType
)
from collie.contracts.mlflow import MLFlowComponentABC
from collie.core.models import (
    TransformerPayload,
    TransformerArtifact
)
from collie._common.decorator import type_checker
from collie._common.exceptions import TransformerError


class Transformer(EventHandler, MLFlowComponentABC):

    def __init__(
        self,
        description: Optional[str] = None,
        tags: Optional[dict] = None
    ) -> None:
        """
        Initializes the Transformer component.

        Args:
            description (Optional[str], optional): Description for the MLflow run. Defaults to None.
            tags (Optional[dict], optional): Tags to associate with the MLflow run. Defaults to None.
        """
        super().__init__()
        self.description = description
        self.tags = tags or {"component": "Transformer"}
    
    def run(self, event: Event) -> Event:
        """
        Run the transformer component.

        This method starts a new MLflow run, transforms the input data,
        logs metrics, and sets the outputs.
        """
        with self.start_run(
            tags=self.tags,
            run_name="Transformer",
            log_system_metrics=True,
            nested=True,
            description=self.description
        ) as run:
            try:
                transformer_event = self._handle(event)

                transformer_payload = self._transformer_payload(transformer_event)
                event_type = EventType.DATA_READY

                data = transformer_payload.model_dump()
                for data_type, data in data.items():

                    if data is not None:
                        source = self.artifact_uri(run, data_type)
                        event.context.set(
                            f"{data_type}_uri",
                            source
                        )
                        
                        self.log_pd_data(
                            data=data, 
                            context=data_type,
                            source=source
                        )

                return Event(
                    type=event_type,
                    payload=transformer_payload,
                    context=event.context
                )

            except Exception as e:
                raise TransformerError(f"Transformer failed with error: {e}")
        
    @type_checker((TransformerPayload,) , 
        "TransformerPayload must be of type TransformerPayload."
    )    
    def _transformer_payload(self, event: Event) -> TransformerPayload: 

        return event.payload
    
    def artifact_uri(self, run, data_type) -> str:
         return f"{run.info.artifact_uri}/{TransformerArtifact().model_dump()[data_type]}" 


from typing import Optional, Dict, Union
from collections import OrderedDict

import torch


def load_pytorch_model_checkpoint(
    ckpt_path: str, 
    model: torch.nn.Module,
    optimizer: Optional[torch.optim.Optimizer] = None,
    lr_scheduler: Optional[torch.optim.lr_scheduler.LRScheduler] = None,
    is_ddp_model: bool = False
) -> Dict[str, Union[torch.nn.Module, torch.optim.Optimizer, torch.optim.lr_scheduler._LRScheduler, int]]:
    """
    Loads a PyTorch model checkpoint and optionally restores the optimizer and learning rate scheduler states.

    Args:
        ckpt_path (str): Path to the checkpoint file.
        model (torch.nn.Module): The model to load the state dict into.
        optimizer (Optional[torch.optim.Optimizer], optional): Optimizer to restore its state. Default is None.
        lr_scheduler (Optional[torch.optim.lr_scheduler._LRScheduler], optional): Learning rate scheduler to restore its state. Default is None.
        is_ddp_model (bool, optional): Whether the model is wrapped in DistributedDataParallel. Default is False.

    Returns:
        Dict[str, Union[torch.nn.Module, torch.optim.Optimizer, torch.optim.lr_scheduler._LRScheduler, int]]: 
            A dictionary containing the restored model, optimizer, learning rate scheduler (if any), and the epoch.
    """

    res_dict = dict()

    checkpoint = torch.load(ckpt_path)
    state_dict = checkpoint['model_state_dict']
    
    # Handle DDP (DistributedDataParallel) model
    if is_ddp_model:
        new_state_dict = OrderedDict()
        for k, v in state_dict.items():
            # remove 'module.' of DataParallel/DistributedDataParallel
            name = k[7:] 
            new_state_dict[name] = v
        state_dict = new_state_dict

    model.load_state_dict(state_dict, strict=False)
    res_dict["model"] = model.eval()
    res_dict["epoch"] = checkpoint.get("epoch", "")

    if optimizer :
        optimizer.load_state_dict(checkpoint['optimizer_state_dict'])
        res_dict["optimizer"] = optimizer

    if 'lr_scheduler_state_dict' in checkpoint and lr_scheduler:
        lr_scheduler.load_state_dict(checkpoint['lr_scheduler_state_dict']) 
        res_dict["lr_scheduler"] = lr_scheduler

    return res_dict


from typing import Any
import warnings


# Import with better error handling
try:
    import mlflow.sklearn
    import sklearn.base
    SKLEARN_AVAILABLE = True
except ImportError:
    SKLEARN_AVAILABLE = False
    warnings.warn("scikit-learn not available. SklearnFlavorHandler will be disabled.")

try:
    import mlflow.xgboost
    import xgboost as xgb
    XGBOOST_AVAILABLE = True
except ImportError:
    XGBOOST_AVAILABLE = False
    warnings.warn("XGBoost not available. XGBoostFlavorHandler will be disabled.")

try:
    import mlflow.pytorch
    import torch.nn as nn
    PYTORCH_AVAILABLE = True
except ImportError:
    PYTORCH_AVAILABLE = False
    warnings.warn("PyTorch not available. PyTorchFlavorHandler will be disabled.")

try:
    import mlflow.lightgbm
    import lightgbm as lgb
    LIGHTGBM_AVAILABLE = True
except (ImportError, Exception):
    LIGHTGBM_AVAILABLE = False
    lgb = None
    warnings.warn("LightGBM not available. LightGBMFlavorHandler will be disabled.")

try:
    import mlflow.transformers
    from transformers import PreTrainedModel
    TRANSFORMERS_AVAILABLE = True
except ImportError:
    TRANSFORMERS_AVAILABLE = False
    warnings.warn("Transformers not available. TransformersFlavorHandler will be disabled.")

from collie._common.mlflow_model_io.base_flavor_handler import FlavorHandler
from collie.core.enums.ml_models import ModelFlavor
from collie._common.exceptions import ModelFlavorError


class SklearnFlavorHandler(FlavorHandler):
    """Handler for scikit-learn models."""
    
    def can_handle(self, model: Any) -> bool:
        if not SKLEARN_AVAILABLE:
            return False
        return isinstance(model, sklearn.base.BaseEstimator)

    def flavor(self) -> ModelFlavor:
        return ModelFlavor.SKLEARN

    def log_model(self, model: Any, name: str, **kwargs: Any) -> None:
        if not SKLEARN_AVAILABLE:
            raise ModelFlavorError(
                "scikit-learn is not available. Please install it to log sklearn models.",
                flavor="sklearn"
            )
        try:
            mlflow.sklearn.log_model(sk_model=model, artifact_path=name, **kwargs)
        except Exception as e:
            raise ModelFlavorError(
                f"Failed to log sklearn model: {str(e)}",
                flavor="sklearn",
                details={"model_type": type(model).__name__, "artifact_name": name}
            ) from e

    def load_model(self, model_uri: str) -> Any:
        if not SKLEARN_AVAILABLE:
            raise ModelFlavorError(
                "scikit-learn is not available. Please install it to load sklearn models.",
                flavor="sklearn"
            )
        try:
            return mlflow.sklearn.load_model(model_uri)
        except Exception as e:
            raise ModelFlavorError(
                f"Failed to load sklearn model: {str(e)}",
                flavor="sklearn",
                details={"model_uri": model_uri}
            ) from e
    

class XGBoostFlavorHandler(FlavorHandler):
    """Handler for XGBoost models."""
    
    def can_handle(self, model: Any) -> bool:
        if not XGBOOST_AVAILABLE:
            return False
        return isinstance(model, (xgb.Booster, xgb.XGBModel))

    def flavor(self) -> ModelFlavor:
        return ModelFlavor.XGBOOST

    def log_model(self, model: Any, name: str, **kwargs: Any) -> None:
        if not XGBOOST_AVAILABLE:
            raise ModelFlavorError(
                "XGBoost is not available. Please install it to log XGBoost models.",
                flavor="xgboost"
            )
        try:
            mlflow.xgboost.log_model(xgb_model=model, artifact_path=name, **kwargs)
        except Exception as e:
            raise ModelFlavorError(
                f"Failed to log XGBoost model: {str(e)}",
                flavor="xgboost",
                details={"model_type": type(model).__name__, "artifact_name": name}
            ) from e

    def load_model(self, model_uri: str) -> Any:
        if not XGBOOST_AVAILABLE:
            raise ModelFlavorError(
                "XGBoost is not available. Please install it to load XGBoost models.",
                flavor="xgboost"
            )
        try:
            return mlflow.xgboost.load_model(model_uri)
        except Exception as e:
            raise ModelFlavorError(
                f"Failed to load XGBoost model: {str(e)}",
                flavor="xgboost",
                details={"model_uri": model_uri}
            ) from e


class PyTorchFlavorHandler(FlavorHandler):
    """Handler for PyTorch models."""
    
    def can_handle(self, model: Any) -> bool:
        if not PYTORCH_AVAILABLE:
            return False
        return isinstance(model, nn.Module)

    def flavor(self) -> ModelFlavor:
        return ModelFlavor.PYTORCH

    def log_model(self, model: Any, name: str, **kwargs: Any) -> None:
        if not PYTORCH_AVAILABLE:
            raise ModelFlavorError(
                "PyTorch is not available. Please install it to log PyTorch models.",
                flavor="pytorch"
            )
        try:
            mlflow.pytorch.log_model(pytorch_model=model, artifact_path=name, **kwargs)
        except Exception as e:
            raise ModelFlavorError(
                f"Failed to log PyTorch model: {str(e)}",
                flavor="pytorch",
                details={"model_type": type(model).__name__, "artifact_name": name}
            ) from e

    def load_model(self, model_uri: str) -> Any:
        if not PYTORCH_AVAILABLE:
            raise ModelFlavorError(
                "PyTorch is not available. Please install it to load PyTorch models.",
                flavor="pytorch"
            )
        try:
            return mlflow.pytorch.load_model(model_uri)
        except Exception as e:
            raise ModelFlavorError(
                f"Failed to load PyTorch model: {str(e)}",
                flavor="pytorch",
                details={"model_uri": model_uri}
            ) from e
    

class LightGBMFlavorHandler(FlavorHandler):
    """Handler for LightGBM models."""
    
    def can_handle(self, model: Any) -> bool:
        if not LIGHTGBM_AVAILABLE or lgb is None:
            return False
        return isinstance(model, (lgb.Booster, lgb.LGBMModel))

    def flavor(self) -> ModelFlavor:
        return ModelFlavor.LIGHTGBM

    def log_model(self, model: Any, name: str, **kwargs: Any) -> None:
        if not LIGHTGBM_AVAILABLE:
            raise ModelFlavorError(
                "LightGBM is not available. Please install it to log LightGBM models.",
                flavor="lightgbm"
            )
        try:
            mlflow.lightgbm.log_model(lgb_model=model, artifact_path=name, **kwargs)
        except Exception as e:
            raise ModelFlavorError(
                f"Failed to log LightGBM model: {str(e)}",
                flavor="lightgbm",
                details={"model_type": type(model).__name__, "artifact_name": name}
            ) from e

    def load_model(self, model_uri: str) -> Any:
        if not LIGHTGBM_AVAILABLE:
            raise ModelFlavorError(
                "LightGBM is not available. Please install it to load LightGBM models.",
                flavor="lightgbm"
            )
        try:
            return mlflow.lightgbm.load_model(model_uri)
        except Exception as e:
            raise ModelFlavorError(
                f"Failed to load LightGBM model: {str(e)}",
                flavor="lightgbm",
                details={"model_uri": model_uri}
            ) from e
    

class TransformersFlavorHandler(FlavorHandler):
    """Handler for Hugging Face Transformers models."""
    
    def can_handle(self, model: Any) -> bool:
        if not TRANSFORMERS_AVAILABLE:
            return False
        return isinstance(model, PreTrainedModel)

    def flavor(self) -> ModelFlavor:
        return ModelFlavor.TRANSFORMERS

    def log_model(self, model: Any, name: str, **kwargs: Any) -> None:
        if not TRANSFORMERS_AVAILABLE:
            raise ModelFlavorError(
                "Transformers is not available. Please install it to log Transformers models.",
                flavor="transformers"
            )
        try:
            mlflow.transformers.log_model(transformers_model=model, artifact_path=name, **kwargs)
        except Exception as e:
            raise ModelFlavorError(
                f"Failed to log Transformers model: {str(e)}",
                flavor="transformers",
                details={"model_type": type(model).__name__, "artifact_name": name}
            ) from e

    def load_model(self, model_uri: str) -> Any:
        if not TRANSFORMERS_AVAILABLE:
            raise ModelFlavorError(
                "Transformers is not available. Please install it to load Transformers models.",
                flavor="transformers"
            )
        try:
            return mlflow.transformers.load_model(model_uri)
        except Exception as e:
            raise ModelFlavorError(
                f"Failed to load Transformers model: {str(e)}",
                flavor="transformers",
                details={"model_uri": model_uri}
            ) from e


from typing import (
    Dict, 
    Any,
    List,
    Optional
)

from collie.contracts.event import (
    Event, 
    EventHandler, 
    EventType
)
from collie.contracts.mlflow import MLFlowComponentABC
from collie._common.decorator import type_checker
from collie.core.models import (
    EvaluatorArtifact,
    EvaluatorPayload
)
from collie._common.exceptions import EvaluatorError

class Evaluator(EventHandler, MLFlowComponentABC):
    
    def __init__(
        self, 
        description: Optional[str] = None,
        tags: Optional[Dict[str, str]] = None
    ) -> None:
        
        """
        Initializes the Evaluator.

        Args:
            description (Optional[str], optional): Description for the MLflow run. Defaults to None.
            tags (Optional[Dict[str, str]], optional): Tags to associate with the MLflow run. Defaults to None.

        """
        super().__init__()
        self._registered_model_name = None
        self.description = description
        self.tags = tags or {"component": "Evaluator"}

        self.model_uri = None
        self.metrics = None
    
    @property
    def registered_model_name(self) -> str:
        if not self._registered_model_name:
            raise EvaluatorError("Registered model name is not set.")
        return self._registered_model_name
    
    @registered_model_name.setter
    def registered_model_name(self, name: str) -> None:
        self._registered_model_name = name
        
    def run(self, event: Event) -> Event:
       
        with self.start_run(
            tags=self.tags,
            run_name="Evaluator",
            log_system_metrics=True,
            description=self.description,
            nested=True,
        ) as run:
            
            try:
                evaluator_event = self._handle(event)
                payload = self._get_evaluator_payload(evaluator_event)

                self.metrics: List[Dict[str, Any]] = payload.metrics
                self.model_uri = event.context.get('model_uri')
        
                self._log_metrics()
                self._log_summary(payload)

                event.context.set("evaluator_report_uri", self.artifact_uri(run))

                if self.experiment_is_better(payload):
                    event.context.set("pass_evaluation", True)
                    self.mlflow.log_param("evaluation_result", "passed")
                else:
                    event.context.set("pass_evaluation", False)
                    self.mlflow.log_param("evaluation_result", "failed")
               
            except Exception as e:
                raise EvaluatorError(f"Evaluator failed: {e}") from e

            return Event(
                type=EventType.EVALUATION_DONE,
                payload=payload,
                context=event.context,
            )

    @staticmethod
    def experiment_is_better(
        payload: EvaluatorPayload
    ) -> bool:
        """
        Decide if experiment model is better based on score direction.
        """
        return payload.is_better_than_production
    
    def _flatten_metrics(
        self,
        metrics: List[Dict[str, Any]]
    ) -> Dict[str, Any]:
        """
        Flatten list of metric dictionaries into a single dictionary.
        Useful for logging and analysis.
        """
        flattened = {}
        for metric_dict in metrics:
            flattened.update(metric_dict)
        return flattened

    def _log_summary(
        self, 
        payload: EvaluatorPayload
    ) -> None:
        """
        Generate summary statistics from metrics.
        """
        flattened = self._flatten_metrics(payload.metrics)
        
        summary = {
            "total_metrics": len(flattened),
            "is_better": payload.is_better_than_production,
            "metrics": flattened
        }
        self.mlflow.log_dict(
            dictionary=summary, 
            artifact_file=f"{EvaluatorArtifact().report}"
        )
    
    def _log_metrics(self) -> None:
        """
        Log individual metrics to MLflow.
        """
        for metric_dict in self.metrics:
            for metric_name, metric_value in metric_dict.items():
                self.mlflow.log_metric(metric_name, metric_value)
    
    @type_checker((EvaluatorPayload,), "EvaluatorPayload must be of type EvaluatorPayload.")
    def _get_evaluator_payload(self, event: Event) -> EvaluatorPayload: 
        return event.payload
    
    def artifact_uri(self, run) -> str:
        return f"{run.info.artifact_uri}/{EvaluatorArtifact().report}"


from typing import Optional
from glob import glob
import os
import shutil

import torch

from collie.helper.pytorch.callback.callback import Callback
from collie._common.utils import get_logger
from collie import trainer

logger = get_logger()

class ModelCheckpoint(Callback):

    def __init__(self, topk_checkpoints: int):
        super().__init__()
        self.topk_checkpoints = topk_checkpoints
        self._best_checkpoints = [] # save as (loss, epoch_idx, checkpoint_path)
        
        self.parent_dir = "./.checkpoint/"
        if os.path.exists(self.parent_dir):
            shutil.rmtree(self.parent_dir)
            logger.info(f"Directory {self.parent_dir} has been removed.")

    def on_epoch_end(
        self,
        trainer: "trainer.PytorchTrainer", 
        epoch_step: int, 
        epoch_train_loss: float, 
        epoch_val_loss: Optional[float]
    ) -> None:
        
        current_loss = epoch_val_loss if epoch_val_loss else epoch_train_loss
        if self._should_save_checkpoint(current_loss):
            self._save_checkpoint(
                trainer=trainer,
                epoch_step=epoch_step,
                epoch_loss=epoch_train_loss,
                loss_for_ckpt=current_loss
            )
    def on_train_end(self, trainer):
        for ckpt in glob(f"{self.parent_dir}/*.pt"):
            trainer.log_artifact(ckpt, "checkpoints")

    def _should_save_checkpoint(self, current_loss: float) -> bool:
        """
        Returns a boolean indicating if the current checkpoint should be saved.

        The current checkpoint should be saved if the topk_checkpoints list is not full.
        Otherwise, the current checkpoint should be saved if its loss is better than the
        loss of the worst checkpoint in the topk_checkpoints list.

        Args:
            current_loss (float): The loss of the current checkpoint.

        Returns:
            bool: True if the current checkpoint should be saved, False otherwise.
        """
        if len(self._best_checkpoints) < self.topk_checkpoints:
            # If the topk_checkpoints list is not full, save the current checkpoint.
            return True

        worst_loss = max(self._best_checkpoints, key=lambda x: x[0])[0]
        # If the current checkpoint has a better loss than the worst checkpoint in the
        # topk_checkpoints list, save the current checkpoint.
        return current_loss < worst_loss
    
    def _save_checkpoint(
        self,
        trainer: "trainer.PytorchTrainer", 
        epoch_step: int, 
        epoch_loss: float,
        loss_for_ckpt: float
    ) -> None:

        checkpoint_path = self.parent_dir + f'model_epoch{epoch_step}.pt'

        if not os.path.exists(self.parent_dir):
            os.makedirs(self.parent_dir)

        # https://zhuanlan.zhihu.com/p/136902153
        checkpoint = {
            "model_state_dict": trainer.model.state_dict(),
            "optimizer_state_dict": trainer.optimizer.state_dict(),
            'scheduler_state_dict': trainer.lr_scheduler.state_dict() if trainer.lr_scheduler else None,
            "epoch": epoch_step,
            "loss": epoch_loss
        }

        if len(self._best_checkpoints) < self.topk_checkpoints:
            self._best_checkpoints.append((loss_for_ckpt, epoch_step, checkpoint_path))
        else:
            worst_idx, (worst_loss, _, worst_path) = max(enumerate(self._best_checkpoints), key=lambda x: x[1][0])

            os.remove(worst_path)  
            self._best_checkpoints[worst_idx] = (loss_for_ckpt, epoch_step, checkpoint_path)

        self._best_checkpoints.sort(key=lambda x: x[0])
        torch.save(checkpoint, f'{checkpoint_path}')


from typing import Optional

from collie.helper.pytorch.callback.callback import Callback
from collie import trainer
from collie._common.utils import get_logger

logger = get_logger()


class EarlyStopping(Callback):

    def __init__(
        self, 
        patience_on_epoch: int, 
        delta: float = 0.0, 
        monitor: str = "val_loss"
    ):
        super().__init__()

        self.patience_on_epoch = patience_on_epoch
        self.delta = delta
        self.monitor = monitor
        self.best_score = float('inf')  
        self.wait = 0  
        self.stopped_epoch = 0  
        self.early_stop = False 

    def on_epoch_end(
        self, 
        epoch_step: int, 
        epoch_train_loss: float, 
        epoch_val_loss: Optional[float], 
        trainer: "trainer.PytorchTrainer"
    ) -> None:
        
        if self.monitor == "val_loss" and epoch_val_loss is not None:
            score = epoch_val_loss
        elif self.monitor == "train_loss":
            score = epoch_train_loss
        else:
            return  

       
        if self.best_score - score > self.delta:  
            self.best_score = score
            self.wait = 0  
        else:
            self.wait += 1  

        if self.wait >= self.patience_on_epoch:
            self.early_stop = True
            self.stopped_epoch = epoch_step
            trainer.should_stop = True
            logger.info(f"Epoch {epoch_step}: Early stopping triggered (patience {self.patience_on_epoch} epochs).")


from typing import (
    Optional,
    Dict,
)

from collie.contracts.orchestrator import OrchestratorBase
from collie.core.enums.components import CollieComponentType
from collie._common.utils import get_logger

logger = get_logger()


class Orchestrator(OrchestratorBase):

    def __init__(
        self,
        components: CollieComponentType,
        tracking_uri: str,
        registered_model_name: str,
        mlflow_tags: Optional[Dict[str, str]] = None,
        experiment_name: Optional[str] = None,
        description: Optional[str] = None,
    ) -> None:
        super().__init__(
            components=components,
            tracking_uri=tracking_uri,
            mlflow_tags=mlflow_tags,
            experiment_name=experiment_name,
            description=description
        )
        self.registered_model_name = registered_model_name

    def orchestrate_pipeline(self) -> None:
        """Run the pipeline sequentially without Airflow."""

        logger.info("Pipeline started.")
        incoming_event = self.initialize_event()

        for idx, component in enumerate(self.components):
            logger.info(f"Running component {idx}: {type(component).__name__}")
            component.mlflow_config = self.mlflow_config
            if hasattr(component, "_registered_model_name"):
                component.registered_model_name = self.registered_model_name
        
            incoming_event = component.run(incoming_event)

            logger.info(f"Component {idx} finished: {type(component).__name__}")


from abc import ABC, abstractmethod
from typing import Any


class FlavorHandler(ABC):

    @abstractmethod
    def can_handle(self, model: Any) -> bool:
        raise NotImplementedError

    @abstractmethod
    def flavor(self):
        raise NotImplementedError

    @abstractmethod
    def log_model(
        self, 
        model: Any, 
        name: str, 
        **kwargs: Any
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def load_model(self, model_uri: str) -> Any:
        raise NotImplementedError


from enum import Enum


class ModelFlavor(str, Enum):
    SKLEARN = "sklearn"
    XGBOOST = "xgboost"
    LIGHTGBM = "lightgbm"
    PYTORCH = "pytorch"
    TRANSFORMERS = "transformers"


class MLflowModelStage(str, Enum):
    
    NONE = "None"
    STAGING = "Staging"
    PRODUCTION = "Production"
    ARCHIVED = "Archived"
    
    def __str__(self) -> str:
        return self.value


from typing import (
    Dict, 
    Any,
    Optional,
    TypeVar
)
from abc import abstractmethod, ABC
from pydantic import (
    Field, 
    BaseModel, 
    ConfigDict
)
from enum import Enum, auto

from collie._common.decorator import type_checker


class PipelineContext:
    def __init__(self, data: Optional[Dict[str, Any]] = None):
        self.data = data or {}

    def get(self, key: str, default=None):
        return self.data.get(key, default)

    def set(self, key: str, value: Any):
        self.data[key] = value

    def to_dict(self):
        return self.data

class EventType(Enum):
    INITIALIZE = auto()
    DATA_READY = auto()
    TRAINING_DONE = auto()
    TUNING_DONE = auto()
    EVALUATION_DONE = auto()
    PUSHER_DONE = auto()
    ERROR = auto()


P = TypeVar("P")
class Event(BaseModel):
    type: Optional[EventType] = None
    payload: P
    context: PipelineContext = Field(default_factory=PipelineContext)

    model_config = ConfigDict(arbitrary_types_allowed=True, populate_by_name=True)


class EventHandler(ABC):
    
    @abstractmethod
    def handle(self, event: Event) -> Event:
        """
        Handle the incoming event and return a new event.
        
        This method must be implemented by all concrete event handlers.
        It should process the event payload and return a new event with
        the appropriate type and payload for the next component.
        
        Args:
            event (Event): The incoming event to process
            
        Returns:
            Event: A new event with processed payload
            
        Raises:
            NotImplementedError: If not implemented by concrete class
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} must implement the 'handle' method. "
            f"This method should process the incoming event and return a new event "
            f"with the appropriate payload for the next pipeline component."
        )
    
    @type_checker((Event,), "The return type of *handle* method must be 'Event'.")
    def _handle(self, event: Event) -> Event:
        """Internal wrapper for handle method with type checking."""
        return self.handle(event)



class CollieBaseException(Exception):
    """Base exception for all Collie framework errors."""
    
    def __init__(self, message: str, component: str = None, details: dict = None):
        self.message = message
        self.component = component or self.__class__.__name__.replace('Error', '')
        self.details = details or {}
        
        detailed_message = f"[{self.component}] {message}"
        if self.details:
            detailed_message += f" Details: {self.details}"
        
        super().__init__(detailed_message)


class MLflowConfigurationError(CollieBaseException):
    """Raised when MLflow configuration is invalid."""
    
    def __init__(self, message: str, config_param: str = None, **kwargs):
        details = kwargs.get('details', {})
        if config_param:
            details['config_parameter'] = config_param
        super().__init__(message, component="MLflow Config", details=details)


class MLflowOperationError(CollieBaseException):
    """Raised when MLflow operations fail."""
    
    def __init__(self, message: str, operation: str = None, **kwargs):
        details = kwargs.get('details', {})
        if operation:
            details['operation'] = operation
        super().__init__(message, component="MLflow Operation", details=details)


class OrchestratorError(CollieBaseException):
    """Raised for errors in the orchestrator process."""
    
    def __init__(self, message: str, pipeline_stage: str = None, **kwargs):
        details = kwargs.get('details', {})
        if pipeline_stage:
            details['pipeline_stage'] = pipeline_stage
        super().__init__(message, component="Orchestrator", details=details)


class TransformerError(CollieBaseException):
    """Raised when data transformation fails."""
    
    def __init__(self, message: str, data_type: str = None, **kwargs):
        details = kwargs.get('details', {})
        if data_type:
            details['data_type'] = data_type
        super().__init__(message, component="Transformer", details=details)


class TrainerError(CollieBaseException):
    """Raised when model training fails."""
    
    def __init__(self, message: str, model_type: str = None, **kwargs):
        details = kwargs.get('details', {})
        if model_type:
            details['model_type'] = model_type
        super().__init__(message, component="Trainer", details=details)


class TunerError(CollieBaseException):
    """Raised when hyperparameter tuning fails."""
    
    def __init__(self, message: str, tuning_method: str = None, **kwargs):
        details = kwargs.get('details', {})
        if tuning_method:
            details['tuning_method'] = tuning_method
        super().__init__(message, component="Tuner", details=details)


class EvaluatorError(CollieBaseException):
    """Raised when model evaluation fails."""
    
    def __init__(self, message: str, metric: str = None, **kwargs):
        details = kwargs.get('details', {})
        if metric:
            details['metric'] = metric
        super().__init__(message, component="Evaluator", details=details)


class PusherError(CollieBaseException):
    """Raised when model pushing/deployment fails."""
    
    def __init__(self, message: str, deployment_target: str = None, **kwargs):
        details = kwargs.get('details', {})
        if deployment_target:
            details['deployment_target'] = deployment_target
        super().__init__(message, component="Pusher", details=details)


class ModelFlavorError(CollieBaseException):
    """Raised when model flavor operations fail."""
    
    def __init__(self, message: str, flavor: str = None, **kwargs):
        details = kwargs.get('details', {})
        if flavor:
            details['flavor'] = flavor
        super().__init__(message, component="Model Flavor", details=details)


import tempfile
import threading
from abc import ABC, abstractmethod
from typing import (
    Any, 
    Optional, 
    Dict, 
    Literal, 
    List,
    Union,
    Generator,
    overload
)
from contextlib import contextmanager

import mlflow
import mlflow.data
import pandas as pd
from mlflow.tracking import MlflowClient
from mlflow import ActiveRun
from mlflow.exceptions import MlflowException

from collie._common.utils import get_logger
from collie._common.mlflow_model_io.model_io import MLflowModelIO
from collie._common.exceptions import (
    MLflowConfigurationError, 
    MLflowOperationError
)
from collie.core.enums.ml_models import (
    MLflowModelStage,
    ModelFlavor
)


logger = get_logger()


class MLflowConfig:
    """Singleton class to manage MLflow configuration."""
    
    _instance = None
    _lock = threading.Lock()
    
    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super().__new__(cls)
                    cls._instance._initialized = False
        return cls._instance
    
    def __init__(
        self,
        tracking_uri: str,
        experiment_name: str,
    ) -> None:
        
        if not self._initialized:
            self.tracking_uri = tracking_uri
            self.experiment_name = experiment_name
            self.mlflow_client = MlflowClient(tracking_uri=tracking_uri)
            self._initialized = True
    
    def configure(self) -> None:
        """Configure the singleton with MLflow settings."""
        mlflow.set_tracking_uri(self.tracking_uri)
        mlflow.set_experiment(self.experiment_name)


class _MLflowModelManager:
    """Handles MLflow model management operations."""
    
    def __init__(
        self, 
        mlflow_client: MlflowClient
    ) -> None:
        
        self._mlflow_client = mlflow_client
        self._model_io = MLflowModelIO(mlflow_client)
    
    def log_model(self, model: Any, name: Optional[str] = None) -> None:
        
        """
        Logs a model with MLflow.

        Args:
            model (Any): The model to log.
            name (Optional[str], optional): The name to give the logged model. Defaults to None.

        Raises:
            MLflowOperationError: If logging the model fails.
        """
        try:
            self._model_io.log_model(model, name)
            logger.info(f"Logged model: {name or 'unnamed'}")
        except Exception as e:
            raise MLflowOperationError(f"Failed to log model '{name}': {e}") from e
    
    def load_model(
        self, 
        flavor: ModelFlavor,
        model_uri: Optional[str] = None
    ) -> Any:
        
        try:
            active_run = mlflow.active_run()
            if active_run is None:
                raise MLflowOperationError("No active run found")
            
            model = self._model_io.load_model(flavor, model_uri)
            return model
        except Exception as e:
            raise MLflowOperationError(f"Failed to load model '{model_uri}': {e}") from e
    
    def load_latest_model(
            self, 
            model_name: str, 
            stage: MLflowModelStage,
            flavor: ModelFlavor
        ) -> Any:
      
        """
        Loads the latest version of a model from a given stage.

        Args:
            model_name (str): The name of the model to load.
            stage (MLflowModelStage): The stage from which to load the model.
            flavor (ModelFlavor): The flavor of the model to load.

        Returns:
            Any: The loaded model if found, otherwise None.

        Raises:
            MLflowOperationError: If loading the model fails.
        """
        try:
            latest_versions = self._mlflow_client.get_latest_versions(model_name, stages=[stage])
            if not latest_versions:
                logger.warning(f"No model found in stage '{stage}' for model '{model_name}'")
                return None
            
            latest = latest_versions[0]  
            model_uri = latest.source 
        except Exception as e:
            logger.error(f"Failed to get latest model version for '{model_name}' in stage '{stage}': {e}")
            return None
            
        return self.load_model(flavor=flavor, model_uri=model_uri)

    def register_model(
        self, 
        model_name: str, 
        model_uri: str
    ) -> int:
        
        """
        Registers a model with MLflow.

        Args:
            model_name (str): The name to give the registered model.
            model_uri (str): The URI of the model to register.

        Returns:
            int: The version number of the registered model.

        Raises:
            MLflowOperationError: If registering the model fails.
        """
        
        try:
            registered_model = mlflow.register_model(model_uri, model_name)
            logger.info(f"Registered model '{model_name}' version {registered_model.version}")
            return registered_model.version
        except MlflowException as e:
            raise MLflowOperationError(
                f"Failed to register model '{model_name}' with URI '{model_uri}': {e}"
            ) from e

    def transition_model_version(
        self,
        registered_model_name: str,
        version: str,
        desired_stage: str,
        archive_existing_versions_at_stage: bool = False,
    ) -> None:
        
        """
        Transitions a model version from one stage to another.

        Args:
            registered_model_name (str): The name of the registered model to transition.
            version (str): The version of the model to transition.
            desired_stage (str): The desired stage to transition the model to.
            archive_existing_versions_at_stage (bool, optional): Whether to archive existing versions at the target stage. Defaults to False.

        Raises:
            MLflowOperationError: If transitioning the model version fails.
        """
        try:
            self._mlflow_client.transition_model_version_stage(
                name=registered_model_name,
                version=version,
                stage=desired_stage,
                archive_existing_versions=archive_existing_versions_at_stage,
            )
            logger.info(f"Transitioned model '{registered_model_name}' v{version} to {desired_stage}")
        except MlflowException as e:
            raise MLflowOperationError(
                f"Failed to transition model '{registered_model_name}' v{version} to {desired_stage}: {e}"
            ) from e
        
    def get_latest_version(
        self,
        model_name: str,
        stages
    ) -> Optional[str]:
       
        """
        Retrieves the latest version number of a model in the specified stages.

        Args:
            model_name (str): The name of the model to retrieve the latest version for.
            stages (List[MLflowModelStage], optional): The stages in which to search for the latest version.

        Returns:
            Optional[str]: The latest version number of the model in the specified stages if found, otherwise None.

        Raises:
            MLflowOperationError: If retrieving the latest version fails.
        """
        try:
            latest_versions = self._mlflow_client.get_latest_versions(model_name, stages=stages)
            if not latest_versions:
                logger.warning(f"No versions found for model '{model_name}' in stages {stages}")
                return None

            latest_version = max(latest_versions, key=lambda v: int(v.version))
            return latest_version.version

        except MlflowException as e:
            raise MLflowOperationError(
                f"Failed to get latest version for model '{model_name}': {e}"
            ) from e


class MLFlowComponentABC(ABC):
    """
    Abstract base class for MLflow components with separated concerns.
    """

    def __init__(self) -> None:
        super().__init__()
        self._mlflow_config = None
        self._model_manager = None
        self.mlflow = mlflow

    @abstractmethod
    def run(self, *args, **kwargs) -> Any:
        raise NotImplementedError("Please implement the **run** method.")
    
    @property
    def mlflow_config(self) -> MLflowConfig:
        
        if self._mlflow_config is None:
            raise MLflowConfigurationError("MLflow client not set") 
        return self._mlflow_config
    
    @mlflow_config.setter
    def mlflow_config(self, mlflow_config: MLflowConfig):
        if not isinstance(mlflow_config, MLflowConfig):
            raise MLflowConfigurationError("mlflow_config must be an instance of MLflowConfig")
        self._mlflow_config = mlflow_config

    @property
    def model_manager(self) -> Optional[_MLflowModelManager]:

        if self._model_manager is None:
            mlflow_client = self._mlflow_config.mlflow_client
            self._model_manager = _MLflowModelManager(mlflow_client)
        return self._model_manager

    def log_pd_data(
        self, 
        data: pd.DataFrame, 
        context: str, 
        source: str
    ) -> None:
        if not isinstance(data, pd.DataFrame):
            raise ValueError("Data must be a pandas DataFrame.")
        
        try:
            ds = self.mlflow.data.from_pandas(data, source=source)
            self.mlflow.log_input(ds, context=context)

            with tempfile.NamedTemporaryFile(delete=True, suffix=".csv") as tmp:
                data.to_csv(tmp.name, index=False)
                self.mlflow.log_artifact(tmp.name)

            logger.debug(f"Logged pandas data for context: {context}")
        except Exception as e:
            raise MLflowOperationError(f"Failed to log pandas data: {e}") from e

    def log_model(
        self, 
        model: Any, 
        name: Optional[str] = None
    ) -> None:

        self.model_manager.log_model(model, name)
    
    def load_model(
        self, 
        name: Optional[str] = None
    ) -> Any:
       
        return self.model_manager.load_model(name)
    
    def load_latest_model(
            self, 
            model_name: str,
            flavor: ModelFlavor,
            stage: MLflowModelStage = MLflowModelStage.PRODUCTION,
        ) -> Any:
        
        return self.model_manager.load_latest_model(
            model_name, 
            stage,
            flavor
        ) 

    def register_model(
        self, 
        model_name: str, 
        model_uri: str
    ) -> int:
        
        return self.model_manager.register_model(model_name, model_uri)

    def transition_model_version(
        self, 
        registered_model_name: str, 
        version: str, 
        desired_stage: str, 
        archive_existing_versions_at_stage: bool = False
    ) -> None:
       
        self.model_manager.transition_model_version(
            registered_model_name, 
            version, 
            desired_stage, 
            archive_existing_versions_at_stage
        )
        
    def get_latest_version(
        self, 
        model_name: str, 
        stages: List[Literal["None", "Staging", "Production", "Archived"]]
    ) -> int:
        
        return self.model_manager.get_latest_version(model_name, stages)

    @overload
    def get_experiment(
        self, 
        return_id: Literal[True]
    ) -> Optional[str]: ...

    @overload
    def get_experiment(
        self, 
        return_id: Literal[False] = False
    ) -> Optional[mlflow.entities.Experiment]: ...
    
    def get_experiment(
        self,
        return_id: bool = False
    ) -> Optional[Union[mlflow.entities.Experiment, str]]:

        """
        Retrieves the MLflow experiment corresponding to the configured experiment name.

        Args:
            return_id (bool, optional): If True, returns the experiment ID instead of the experiment object. 
                                        Defaults to False.

        Returns:
            Optional[Union[mlflow.entities.Experiment, str]]: The experiment object or experiment ID if return_id is True, 
                                                                or None if the experiment does not exist.
    """
        experiment_name = self.mlflow_config.experiment_name
        if not experiment_name:
            return None
        try:
            experiment = self.mlflow.get_experiment_by_name(experiment_name)
            if return_id:
                return experiment.experiment_id
            else:
                return experiment
        except MlflowException as e:
            logger.error(f"Failed to get experiment '{experiment_name}': {e}")
            return None
        

    @contextmanager
    def start_run(
        self,
        tags: Optional[Dict[str, str]] = None,
        run_name: Optional[str] = None,
        nested: bool = False,
        log_system_metrics: Optional[bool] = None,
        description: Optional[str] = None,
    ) -> Generator[ActiveRun, None, None]:
        """
        Starts an MLflow run and returns the active run object.

        Args:
            tags (Optional[Dict[str, str]], optional): A dictionary of string key-value pairs to store as run tags. Defaults to None.
            run_name (Optional[str], optional): Name for the run. Defaults to None.
            nested (bool, optional): If True, nested runs are enabled. Defaults to False.
            log_system_metrics (Optional[bool], optional): If True, system metrics are logged. Defaults to None.
            description (Optional[str], optional): A string description for the run. Defaults to None.

        Yields:
            Generator[ActiveRun, None, None]: The active run object.
        Raises:
            MLflowOperationError: If the MLflow run cannot be started.
        """
        try:
            self.mlflow_config.configure()
            experiment_id = self.get_experiment(return_id=True)

            with self.mlflow.start_run(
                experiment_id=experiment_id,
                run_name=run_name,
                nested=nested,
                tags=tags,
                log_system_metrics=log_system_metrics,
                description=description,
            ) as active_run:
                logger.info(f"Started MLflow run: {active_run.info.run_id}")
                yield active_run

        except MlflowException as e:
            raise MLflowOperationError(f"Failed to start MLflow run: {e}") from e